@@ -23,9 +23,10 @@ var LibraryExceptions = {
   //
   // excPtr - Thrown object pointer to wrap. Metadata pointer is calculated from it.
   $ExceptionInfo__docs: '/** @constructor */',
-  $ExceptionInfo__deps: ['__cxa_is_pointer_type',
+  $ExceptionInfo__deps: [
+    '__cxa_is_pointer_type',
 #if EXCEPTION_DEBUG
-    '$ptrToString',
+    '$ptrToString'
 #endif
   ],
   $ExceptionInfo: function(excPtr) {
@@ -48,25 +49,21 @@ var LibraryExceptions = {
       return {{{ makeGetValue('this.ptr', C_STRUCTS.__cxa_exception.exceptionDestructor, '*') }}};
     };
 
-    this.set_refcount = function(refcount) {
-      {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.referenceCount, 'refcount', 'i32') }}};
-    };
-
-    this.set_caught = function (caught) {
+    this.set_caught = function(caught) {
       caught = caught ? 1 : 0;
       {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.caught, 'caught', 'i8') }}};
     };
 
-    this.get_caught = function () {
+    this.get_caught = function() {
       return {{{ makeGetValue('this.ptr', C_STRUCTS.__cxa_exception.caught, 'i8') }}} != 0;
     };
 
-    this.set_rethrown = function (rethrown) {
+    this.set_rethrown = function(rethrown) {
       rethrown = rethrown ? 1 : 0;
       {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.rethrown, 'rethrown', 'i8') }}};
     };
 
-    this.get_rethrown = function () {
+    this.get_rethrown = function() {
       return {{{ makeGetValue('this.ptr', C_STRUCTS.__cxa_exception.rethrown, 'i8') }}} != 0;
     };
 
@@ -78,34 +75,8 @@ var LibraryExceptions = {
       this.set_adjusted_ptr(0);
       this.set_type(type);
       this.set_destructor(destructor);
-      this.set_refcount(0);
-      this.set_caught(false);
-      this.set_rethrown(false);
     }
 
-    this.add_ref = function() {
-#if SHARED_MEMORY
-      Atomics.add(HEAP32, (this.ptr + {{{ C_STRUCTS.__cxa_exception.referenceCount }}}) >> 2, 1);
-#else
-      var value = {{{ makeGetValue('this.ptr', C_STRUCTS.__cxa_exception.referenceCount, 'i32') }}};
-      {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.referenceCount, 'value + 1', 'i32') }}};
-#endif
-    };
-
-    // Returns true if last reference released.
-    this.release_ref = function() {
-#if SHARED_MEMORY
-      var prev = Atomics.sub(HEAP32, (this.ptr + {{{ C_STRUCTS.__cxa_exception.referenceCount }}}) >> 2, 1);
-#else
-      var prev = {{{ makeGetValue('this.ptr', C_STRUCTS.__cxa_exception.referenceCount, 'i32') }}};
-      {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.referenceCount, 'prev - 1', 'i32') }}};
-#endif
-#if ASSERTIONS
-      assert(prev > 0);
-#endif
-      return prev === 1;
-    };
-
     this.set_adjusted_ptr = function(adjustedPtr) {
       {{{ makeSetValue('this.ptr', C_STRUCTS.__cxa_exception.adjustedPtr, 'adjustedPtr', '*') }}};
     };
@@ -131,74 +102,26 @@ var LibraryExceptions = {
     };
   },
 
-  $exception_addRef: function (info) {
-#if EXCEPTION_DEBUG
-    dbg('exception_addRef ' + ptrToString(info.excPtr));
-#endif
-    info.add_ref();
-  },
-
-  $exception_decRef__deps: ['__cxa_free_exception'
-#if EXCEPTION_DEBUG
-    , '$exceptionLast', '$exceptionCaught'
-#endif
-  ],
-  $exception_decRef: function(info) {
-#if EXCEPTION_DEBUG
-    dbg('exception_decRef ' + ptrToString(info.excPtr));
-#endif
-    // A rethrown exception can reach refcount 0; it must not be discarded
-    // Its next handler will clear the rethrown flag and addRef it, prior to
-    // final decRef and destruction here
-    if (info.release_ref() && !info.get_rethrown()) {
-      var destructor = info.get_destructor();
-      if (destructor) {
-        // In Wasm, destructors return 'this' as in ARM
-        {{{ makeDynCall('pp', 'destructor') }}}(info.excPtr);
-      }
-      ___cxa_free_exception(info.excPtr);
-#if EXCEPTION_DEBUG
-      dbg('decref freeing exception ' + [ptrToString(info.excPtr), exceptionLast, 'stack', exceptionCaught]);
-#endif
-    }
-  },
-
-  __cxa_increment_exception_refcount__deps: ['$exception_addRef', '$ExceptionInfo'],
-  __cxa_increment_exception_refcount__sig: 'vp',
-  __cxa_increment_exception_refcount: function(ptr) {
-    if (!ptr) return;
-    exception_addRef(new ExceptionInfo(ptr));
-  },
-
-  __cxa_decrement_exception_refcount__deps: ['$exception_decRef', '$ExceptionInfo'],
-  __cxa_decrement_exception_refcount__sig: 'vp',
-  __cxa_decrement_exception_refcount: function(ptr) {
-    if (!ptr) return;
-    exception_decRef(new ExceptionInfo(ptr));
-  },
-
   // Here, we throw an exception after recording a couple of values that we need to remember
   // We also remember that it was the last exception thrown as we need to know that later.
-  __cxa_throw__sig: 'vppp',
   __cxa_throw__deps: ['$ExceptionInfo', '$exceptionLast', '$uncaughtExceptionCount'],
-  __cxa_throw: function(ptr, type, destructor) {
+  __cxa_throw: (ptr, type, destructor) => {
 #if EXCEPTION_DEBUG
     dbg('__cxa_throw: ' + [ptrToString(ptr), type, ptrToString(destructor)]);
 #endif
     var info = new ExceptionInfo(ptr);
     // Initialize ExceptionInfo content after it was allocated in __cxa_allocate_exception.
     info.init(type, destructor);
-    exceptionLast = ptr;
+    {{{ storeException('exceptionLast', 'ptr') }}}
     uncaughtExceptionCount++;
-    {{{ makeThrow('ptr') }}}
+    {{{ makeThrow('exceptionLast') }}}
   },
 
   // This exception will be caught twice, but while begin_catch runs twice,
   // we early-exit from end_catch when the exception has been rethrown, so
   // pop that here from the caught exceptions.
   __cxa_rethrow__deps: ['$exceptionCaught', '$exceptionLast', '$uncaughtExceptionCount'],
-  __cxa_rethrow__sig: 'v',
-  __cxa_rethrow: function() {
+  __cxa_rethrow: () => {
     var info = exceptionCaught.pop();
     if (!info) {
       abort('no exception to throw');
@@ -215,18 +138,15 @@ var LibraryExceptions = {
     dbg('__cxa_rethrow, popped ' +
       [ptrToString(ptr), exceptionLast, 'stack', exceptionCaught]);
 #endif
-    exceptionLast = ptr;
-    {{{ makeThrow('ptr') }}}
+    {{{ storeException('exceptionLast', 'ptr') }}}
+    {{{ makeThrow('exceptionLast') }}}
   },
 
-  llvm_eh_typeid_for__sig: 'ip',
-  llvm_eh_typeid_for: function(type) {
-    return type;
-  },
+  llvm_eh_typeid_for: (type) => type,
 
-  __cxa_begin_catch__deps: ['$exceptionCaught', '$exception_addRef', '$uncaughtExceptionCount'],
-  __cxa_begin_catch__sig: 'pp',
-  __cxa_begin_catch: function(ptr) {
+  __cxa_begin_catch__deps: ['$exceptionCaught', '__cxa_increment_exception_refcount',
+                            '$uncaughtExceptionCount'],
+  __cxa_begin_catch: (ptr) => {
     var info = new ExceptionInfo(ptr);
     if (!info.get_caught()) {
       info.set_caught(true);
@@ -237,7 +157,7 @@ var LibraryExceptions = {
 #if EXCEPTION_DEBUG
     dbg('__cxa_begin_catch ' + [ptrToString(ptr), 'stack', exceptionCaught]);
 #endif
-    exception_addRef(info);
+    ___cxa_increment_exception_refcount(info.excPtr);
     return info.get_exception_ptr();
   },
 
@@ -245,11 +165,10 @@ var LibraryExceptions = {
   // and free the exception. Note that if the dynCall on the destructor fails
   // due to calling apply on undefined, that means that the destructor is
   // an invalid index into the FUNCTION_TABLE, so something has gone wrong.
-  __cxa_end_catch__deps: ['$exceptionCaught', '$exceptionLast', '$exception_decRef'],
-  __cxa_end_catch__sig: 'v',
-  __cxa_end_catch: function() {
+  __cxa_end_catch__deps: ['$exceptionCaught', '$exceptionLast', '__cxa_decrement_exception_refcount', 'setThrew'],
+  __cxa_end_catch: () => {
     // Clear state flag.
-    _setThrew(0);
+    _setThrew(0, 0);
 #if ASSERTIONS
     assert(exceptionCaught.length > 0);
 #endif
@@ -259,44 +178,36 @@ var LibraryExceptions = {
 #if EXCEPTION_DEBUG
     dbg('__cxa_end_catch popped ' + [info, exceptionLast, 'stack', exceptionCaught]);
 #endif
-    exception_decRef(info);
+    ___cxa_decrement_exception_refcount(info.excPtr);
     exceptionLast = 0; // XXX in decRef?
   },
 
   __cxa_get_exception_ptr__deps: ['$ExceptionInfo'],
-  __cxa_get_exception_ptr__sig: 'pp',
-  __cxa_get_exception_ptr: function(ptr) {
+  __cxa_get_exception_ptr: (ptr) => {
+    var rtn = new ExceptionInfo(ptr).get_exception_ptr();
 #if EXCEPTION_DEBUG
-    dbg('__cxa_get_exception_ptr ' + ptrToString(ptr));
+    dbg('__cxa_get_exception_ptr ' + ptrToString(ptr) + ' -> ' + ptrToString(rtn));
 #endif
-    return new ExceptionInfo(ptr).get_exception_ptr();
+    return rtn;
   },
 
   __cxa_uncaught_exceptions__deps: ['$uncaughtExceptionCount'],
-  __cxa_uncaught_exceptions: function() {
-    return uncaughtExceptionCount;
-  },
+  __cxa_uncaught_exceptions: () => uncaughtExceptionCount,
 
-  __cxa_call_unexpected: function(exception) {
-    err('Unexpected exception thrown, this is not properly supported - aborting');
-#if !MINIMAL_RUNTIME
-    ABORT = true;
-#endif
-    throw exception;
-  },
+  __cxa_call_unexpected: (exception) => abort('Unexpected exception thrown, this is not properly supported - aborting'),
 
-  __cxa_current_primary_exception__deps: ['$exceptionCaught', '$exception_addRef'],
-  __cxa_current_primary_exception: function() {
+  __cxa_current_primary_exception__deps: ['$exceptionCaught', '__cxa_increment_exception_refcount'],
+  __cxa_current_primary_exception: () => {
     if (!exceptionCaught.length) {
       return 0;
     }
     var info = exceptionCaught[exceptionCaught.length - 1];
-    exception_addRef(info);
+    ___cxa_increment_exception_refcount(info.excPtr);
     return info.excPtr;
   },
 
   __cxa_rethrow_primary_exception__deps: ['$ExceptionInfo', '$exceptionCaught', '__cxa_rethrow'],
-  __cxa_rethrow_primary_exception: function(ptr) {
+  __cxa_rethrow_primary_exception: (ptr) => {
     if (!ptr) return;
     var info = new ExceptionInfo(ptr);
     exceptionCaught.push(info);
@@ -313,10 +224,14 @@ var LibraryExceptions = {
   // unwinding using 'if' blocks around each function, so the remaining
   // functionality boils down to picking a suitable 'catch' block.
   // We'll do that here, instead, to keep things simpler.
-  __cxa_find_matching_catch__deps: ['$exceptionLast', '$ExceptionInfo', '__resumeException', '__cxa_can_catch', 'setTempRet0'],
-  //__cxa_find_matching_catch__sig: 'p',
-  __cxa_find_matching_catch: function() {
-    var thrown = exceptionLast;
+  $findMatchingCatch__deps: ['$exceptionLast', '$ExceptionInfo', '__resumeException', '__cxa_can_catch', 'setTempRet0'],
+  $findMatchingCatch: (args) => {
+    var thrown =
+#if EXCEPTION_STACK_TRACES
+      exceptionLast?.excPtr;
+#else
+      exceptionLast;
+#endif
     if (!thrown) {
       // just pass through the null ptr
       setTempRet0(0);
@@ -333,14 +248,15 @@ var LibraryExceptions = {
 
     // can_catch receives a **, add indirection
 #if EXCEPTION_DEBUG
-    dbg("__cxa_find_matching_catch on " + ptrToString(thrown));
+    dbg("findMatchingCatch on " + ptrToString(thrown));
 #endif
     // The different catch blocks are denoted by different types.
     // Due to inheritance, those types may not precisely match the
     // type of the thrown object. Find one which matches, and
     // return the type of the catch block which should be called.
-    for (var i = 0; i < arguments.length; i++) {
-      var caughtType = arguments[i];
+    for (var arg in args) {
+      var caughtType = args[arg];
+
       if (caughtType === 0 || caughtType === thrownType) {
         // Catch all clause matched or exactly the same type is caught
         break;
@@ -348,7 +264,7 @@ var LibraryExceptions = {
       var adjusted_ptr_addr = info.ptr + {{{ C_STRUCTS.__cxa_exception.adjustedPtr }}};
       if (___cxa_can_catch(caughtType, thrownType, adjusted_ptr_addr)) {
 #if EXCEPTION_DEBUG
-        dbg("  __cxa_find_matching_catch found " + [ptrToString(info.get_adjusted_ptr()), caughtType]);
+        dbg("  findMatchingCatch found " + [ptrToString(info.get_adjusted_ptr()), caughtType]);
 #endif
         setTempRet0(caughtType);
         return thrown;
@@ -359,49 +275,48 @@ var LibraryExceptions = {
   },
 
   __resumeException__deps: ['$exceptionLast'],
-  __resumeException__sig: 'vp',
-  __resumeException: function(ptr) {
+  __resumeException: (ptr) => {
 #if EXCEPTION_DEBUG
     dbg("__resumeException " + [ptrToString(ptr), exceptionLast]);
 #endif
-    if (!exceptionLast) { exceptionLast = ptr; }
-    {{{ makeThrow('ptr') }}}
+    if (!exceptionLast) {
+      {{{ storeException('exceptionLast', 'ptr') }}}
+    }
+    {{{ makeThrow('exceptionLast') }}}
   },
 
 #endif
 #if WASM_EXCEPTIONS || !DISABLE_EXCEPTION_CATCHING
   $getExceptionMessageCommon__deps: ['__get_exception_message', 'free', '$withStackSave'],
-  $getExceptionMessageCommon: function(ptr) {
-    return withStackSave(function() {
-      var type_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
-      var message_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
-      ___get_exception_message({{{ to64('ptr') }}}, {{{ to64('type_addr_addr') }}}, {{{ to64('message_addr_addr') }}});
-      var type_addr = {{{ makeGetValue('type_addr_addr', 0, '*') }}};
-      var message_addr = {{{ makeGetValue('message_addr_addr', 0, '*') }}};
-      var type = UTF8ToString(type_addr);
-      _free(type_addr);
-      var message;
-      if (message_addr) {
-        message = UTF8ToString(message_addr);
-        _free(message_addr);
-      }
-      return [type, message];
-    });
-  },
+  $getExceptionMessageCommon: (ptr) => withStackSave(() => {
+    var type_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
+    var message_addr_addr = stackAlloc({{{ POINTER_SIZE }}});
+    ___get_exception_message(ptr, type_addr_addr, message_addr_addr);
+    var type_addr = {{{ makeGetValue('type_addr_addr', 0, '*') }}};
+    var message_addr = {{{ makeGetValue('message_addr_addr', 0, '*') }}};
+    var type = UTF8ToString(type_addr);
+    _free(type_addr);
+    var message;
+    if (message_addr) {
+      message = UTF8ToString(message_addr);
+      _free(message_addr);
+    }
+    return [type, message];
+  }),
 #endif
 #if WASM_EXCEPTIONS
-  $getCppExceptionTag: function() {
+  $getCppExceptionTag: () =>
     // In static linking, tags are defined within the wasm module and are
     // exported, whereas in dynamic linking, tags are defined in library.js in
     // JS code and wasm modules import them.
 #if RELOCATABLE
-    return ___cpp_exception; // defined in library.js
+    ___cpp_exception // defined in library.js
 #else
-    return Module['asm']['__cpp_exception'];
+    wasmExports['__cpp_exception']
 #endif
-  },
+  ,
 
-#if ASSERTIONS
+#if EXCEPTION_STACK_TRACES
   // Throw a WebAssembly.Exception object with the C++ tag with a stack trace
   // embedded. WebAssembly.Exception is a JS object representing a Wasm
   // exception, provided by Wasm JS API:
@@ -409,7 +324,7 @@ var LibraryExceptions = {
   // In release builds, this function is not needed and the native
   // _Unwind_RaiseException in libunwind is used instead.
   __throw_exception_with_stack_trace__deps: ['$getCppExceptionTag', '$getExceptionMessage'],
-  __throw_exception_with_stack_trace: function(ex) {
+  __throw_exception_with_stack_trace: (ex) => {
     var e = new WebAssembly.Exception(getCppExceptionTag(), [ex], {traceStack: true});
     e.message = getExceptionMessage(e);
     // The generated stack trace will be in the form of:
@@ -420,10 +335,12 @@ var LibraryExceptions = {
     //     ...
     //
     // Remove this JS function name, which is in the second line, from the stack
-    // trace.
-    var arr = e.stack.split('\n');
-    arr.splice(1,1);
-    e.stack = arr.join('\n');
+    // trace. Note that .stack does not yet exist in all browsers (see #18828).
+    if (e.stack) {
+      var arr = e.stack.split('\n');
+      arr.splice(1,1);
+      e.stack = arr.join('\n');
+    }
     throw e;
   },
 #endif
@@ -431,7 +348,7 @@ var LibraryExceptions = {
   // Given an WebAssembly.Exception object, returns the actual user-thrown
   // C++ object address in the Wasm memory.
   $getCppExceptionThrownObjectFromWebAssemblyException__deps: ['$getCppExceptionTag', '__thrown_object_from_unwind_exception'],
-  $getCppExceptionThrownObjectFromWebAssemblyException: function(ex) {
+  $getCppExceptionThrownObjectFromWebAssemblyException: (ex) => {
     // In Wasm EH, the value extracted from WebAssembly.Exception is a pointer
     // to the unwind header. Convert it to the actual thrown value.
     var unwind_header = ex.getArg(getCppExceptionTag(), 0);
@@ -439,49 +356,70 @@ var LibraryExceptions = {
   },
 
   $incrementExceptionRefcount__deps: ['__cxa_increment_exception_refcount', '$getCppExceptionThrownObjectFromWebAssemblyException'],
-  $incrementExceptionRefcount: function(ex) {
+  $incrementExceptionRefcount: (ex) => {
     var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
     ___cxa_increment_exception_refcount(ptr);
   },
 
   $decrementExceptionRefcount__deps: ['__cxa_decrement_exception_refcount', '$getCppExceptionThrownObjectFromWebAssemblyException'],
-  $decrementExceptionRefcount: function(ex) {
+  $decrementExceptionRefcount: (ex) => {
     var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
     ___cxa_decrement_exception_refcount(ptr);
   },
 
   $getExceptionMessage__deps: ['$getCppExceptionThrownObjectFromWebAssemblyException', '$getExceptionMessageCommon'],
-  $getExceptionMessage: function(ex) {
+  $getExceptionMessage: (ex) => {
     var ptr = getCppExceptionThrownObjectFromWebAssemblyException(ex);
     return getExceptionMessageCommon(ptr);
   },
 
 #elif !DISABLE_EXCEPTION_CATCHING
   $incrementExceptionRefcount__deps: ['__cxa_increment_exception_refcount'],
-  $incrementExceptionRefcount: function(ptr) {
-    ___cxa_increment_exception_refcount(ptr);
-  },
+  $incrementExceptionRefcount: (ptr) => ___cxa_increment_exception_refcount(ptr),
 
   $decrementExceptionRefcount__deps: ['__cxa_decrement_exception_refcount'],
-  $decrementExceptionRefcount: function(ptr) {
-    ___cxa_decrement_exception_refcount(ptr);
-  },
-  $getExceptionMessage__deps: ['$getExceptionMessageCommon'],
-  $getExceptionMessage: function(ptr) {
-    return getExceptionMessageCommon(ptr);
-  },
+  $decrementExceptionRefcount: (ptr) => ___cxa_decrement_exception_refcount(ptr),
 
+  $getExceptionMessage__deps: ['$getExceptionMessageCommon'],
+  $getExceptionMessage: (ptr) => getExceptionMessageCommon(ptr),
 
 #endif
 };
 
+#if !WASM_EXCEPTIONS
 // In LLVM, exceptions generate a set of functions of form
-// __cxa_find_matching_catch_1(), __cxa_find_matching_catch_2(), etc.  where the
+// __cxa_find_matching_catch_2(), __cxa_find_matching_catch_3(), etc.  where the
 // number specifies the number of arguments.  In Emscripten, route all these to
 // a single function '__cxa_find_matching_catch' that variadically processes all
 // of these functions using JS 'arguments' object.
-addCxaCatch = function(n) {
-  LibraryManager.library['__cxa_find_matching_catch_' + n] = '__cxa_find_matching_catch';
+addCxaCatch = (n) => {
+  const args = [];
+  // Confusingly, the actual number of asrgument is n - 2. According to the llvm
+  // code in WebAssemblyLowerEmscriptenEHSjLj.cpp:
+  // This is because a landingpad instruction contains two more arguments, a
+  // personality function and a cleanup bit, and __cxa_find_matching_catch_N
+  // functions are named after the number of arguments in the original landingpad
+  // instruction.
+  let sig = 'p';
+  for (let i = 0; i < n - 2; i++) {
+    args.push(`arg${i}`);
+    sig += 'p';
+  }
+  const argString = args.join(',');
+  LibraryManager.library[`__cxa_find_matching_catch_${n}__sig`] = sig;
+  LibraryManager.library[`__cxa_find_matching_catch_${n}__deps`] = ['$findMatchingCatch'];
+  LibraryManager.library[`__cxa_find_matching_catch_${n}`] = eval(`(${args}) => findMatchingCatch([${argString}])`);
 };
 
-mergeInto(LibraryManager.library, LibraryExceptions);
+// Add the first 2-5 catch handlers premptively.  Others get added on demand in
+// jsifier.  This is done here primarily so that these symbols end up with the
+// correct deps in the stub library that we pass to wasm-ld.
+// Note: __cxa_find_matching_catch_N function uses N = NumClauses + 2 so
+// __cxa_find_matching_catch_2 is the first such function with zero clauses.
+// See WebAssemblyLowerEmscriptenEHSjLj.cpp.
+for (let i = 2; i < 5; i++) {
+  addCxaCatch(i)
+}
+#endif
+
+addToLibrary(LibraryExceptions);
