@@ -1,7 +1,9 @@
 #pragma once
 
+#include <stddef.h>
 #include <stdint.h>
-#include <emscripten/html5.h>
+#include <emscripten/atomic.h>
+#include <emscripten/em_types.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -88,118 +90,6 @@ void emscripten_wasm_worker_post_function_vdd(emscripten_wasm_worker_t id, void
 void emscripten_wasm_worker_post_function_vddd(emscripten_wasm_worker_t id, void (*funcPtr)(double, double, double) __attribute__((nonnull)), double arg0, double arg1, double arg2);
 void emscripten_wasm_worker_post_function_sig(emscripten_wasm_worker_t id, void *funcPtr __attribute__((nonnull)), const char *sig __attribute__((nonnull)), ...);
 
-#define ATOMICS_WAIT_RESULT_T int
-
-// Numbering dictated by https://github.com/WebAssembly/threads/blob/master/proposals/threads/Overview.md#wait
-#define ATOMICS_WAIT_OK 0
-#define ATOMICS_WAIT_NOT_EQUAL 1
-#define ATOMICS_WAIT_TIMED_OUT 2
-
-#define ATOMICS_WAIT_DURATION_INFINITE -1ll
-
-// Issues the wasm 'memory.atomic.wait32' instruction:
-// If the given memory address contains value 'expectedValue', puts the calling
-// thread to sleep to wait for that address to be notified.
-// Returns one of the ATOMICS_WAIT_* return codes.
-// NOTE: This function takes in the wait value in int64_t nanosecond units. Pass
-// in maxWaitNanoseconds = -1 (or ATOMICS_WAIT_DURATION_INFINITE) to wait
-// infinitely long.
-static __attribute__((always_inline)) ATOMICS_WAIT_RESULT_T emscripten_wasm_wait_i32(int32_t *address __attribute__((nonnull)), int expectedValue, int64_t maxWaitNanoseconds)
-{
-  return __builtin_wasm_memory_atomic_wait32(address, expectedValue, maxWaitNanoseconds);
-}
-
-// Issues the wasm 'memory.atomic.wait64' instruction:
-// If the given memory address contains value 'expectedValue', puts the calling
-// thread to sleep to wait for that address to be notified.
-// Returns one of the ATOMICS_WAIT_* return codes.
-// NOTE: This function takes in the wait value in int64_t nanosecond units. Pass
-// in maxWaitNanoseconds = -1 (or ATOMICS_WAIT_DURATION_INFINITE) to wait
-// infinitely long.
-static __attribute__((always_inline)) ATOMICS_WAIT_RESULT_T emscripten_wasm_wait_i64(int64_t *address __attribute__((nonnull)), int64_t expectedValue, int64_t maxWaitNanoseconds)
-{
-  return __builtin_wasm_memory_atomic_wait64(address, expectedValue, maxWaitNanoseconds);
-}
-
-#define EMSCRIPTEN_NOTIFY_ALL_WAITERS (-1LL)
-
-// Issues the wasm 'memory.atomic.notify' instruction:
-// Notifies the given number of threads waiting on a location.
-// Pass count == EMSCRIPTEN_NOTIFY_ALL_WAITERS to notify all waiters on the
-// given location.
-// Returns the number of threads that were woken up.
-// Note: this function is used to notify both waiters waiting on an i32 and i64
-// addresses.
-static __attribute__((always_inline)) int64_t emscripten_wasm_notify(int32_t *address __attribute__((nonnull)), int64_t count)
-{
-  return __builtin_wasm_memory_atomic_notify(address, count);
-}
-
-#define EMSCRIPTEN_WAIT_ASYNC_INFINITY __builtin_inf()
-
-// Represents a pending 'Atomics.waitAsync' wait operation.
-#define ATOMICS_WAIT_TOKEN_T int32_t
-
-#define EMSCRIPTEN_IS_VALID_WAIT_TOKEN(token) ((token) <= 0)
-
-// Issues the JavaScript 'Atomics.waitAsync' instruction:
-// performs an asynchronous wait operation on the main thread. If the given
-// 'address' contains 'value', issues a deferred wait that will invoke the
-// specified callback function 'asyncWaitFinished' once that address has been
-// notified by another thread.
-// NOTE: Unlike functions emscripten_wasm_wait_i32() and
-// emscripten_wasm_wait_i64() which take in the wait timeout parameter as int64
-// nanosecond units, this function takes in the wait timeout parameter as double
-// millisecond units. See https://github.com/WebAssembly/threads/issues/175 for
-// more information.
-// Pass in maxWaitMilliseconds == EMSCRIPTEN_WAIT_ASYNC_INFINITY
-// (==__builtin_inf()) to wait infinitely long.
-// Returns one of:
-//  - ATOMICS_WAIT_NOT_EQUAL if the waitAsync operation could not be registered
-//    since the memory value did not contain the value 'value'.
-//  - ATOMICS_WAIT_TIMED_OUT if the waitAsync operation timeout parameter was <= 0.
-//  - Any other value: denotes a 'wait token' that can be passed to function
-//    emscripten_atomic_cancel_wait_async() to unregister an asynchronous wait.
-//    You can use the macro EMSCRIPTEN_IS_VALID_WAIT_TOKEN(retval) to check if
-//    this function returned a valid wait token.
-ATOMICS_WAIT_TOKEN_T emscripten_atomic_wait_async(int32_t *address __attribute__((nonnull)),
-                                                  uint32_t value,
-                                                  void (*asyncWaitFinished)(int32_t *address, uint32_t value, ATOMICS_WAIT_RESULT_T waitResult, void *userData) __attribute__((nonnull)),
-                                                  void *userData,
-                                                  double maxWaitMilliseconds);
-
-// Unregisters a pending Atomics.waitAsync operation that was established via a
-// call to emscripten_atomic_wait_async() in the calling thread. Pass in the
-// wait token handle that was received as the return value from the wait
-// function.  Returns EMSCRIPTEN_RESULT_SUCCESS if the cancellation was
-// successful, or EMSCRIPTEN_RESULT_INVALID_PARAM if the asynchronous wait has
-// already resolved prior and the callback has already been called.
-// NOTE: Because of needing to work around issue
-// https://github.com/WebAssembly/threads/issues/176, calling this function has
-// an effect of introducing spurious wakeups to any other threads waiting on the
-// same address that the async wait denoted by the token does. This means that
-// in order to safely use this function, the mechanisms used in any wait code on
-// that address must be written to be spurious wakeup safe. (this is the case
-// for all the synchronization primitives declared in this header, but if you
-// are rolling out your own, you need to be aware of this). If
-// https://github.com/tc39/proposal-cancellation/issues/29 is resolved, then the
-// spurious wakeups can be avoided.
-EMSCRIPTEN_RESULT emscripten_atomic_cancel_wait_async(ATOMICS_WAIT_TOKEN_T waitToken);
-
-// Cancels all pending async waits in the calling thread. Because of
-// https://github.com/WebAssembly/threads/issues/176, if you are using
-// asynchronous waits in your application, and need to be able to let GC reclaim
-// Wasm heap memory when deinitializing an application, you *must* call this
-// function to help the GC unpin all necessary memory.  Otherwise, you can wrap
-// the Wasm content in an iframe and unload the iframe to let GC occur.
-// (navigating away from the page or closing that tab will also naturally
-// reclaim the memory)
-int emscripten_atomic_cancel_all_wait_asyncs(void);
-
-// Cancels all pending async waits in the calling thread to the given memory
-// address.  Returns the number of async waits canceled.
-int emscripten_atomic_cancel_all_wait_asyncs_at_address(int32_t *address __attribute__((nonnull)));
-
 // Sleeps the calling wasm worker for the given nanoseconds. Calling this
 // function on the main thread either results in a TypeError exception
 // (Firefox), or a silent return without waiting (Chrome), see
@@ -389,6 +279,15 @@ ATOMICS_WAIT_TOKEN_T emscripten_condvar_wait_async(emscripten_condvar_t *condvar
 // ("broadcast" operation).
 void emscripten_condvar_signal(emscripten_condvar_t *condvar __attribute__((nonnull)), int64_t numWaitersToSignal);
 
+// Legacy names for emscripten_atomic_wait/notify functions, defined in
+// emscripten/atomic.h
+#define emscripten_wasm_wait_i32 emscripten_atomic_wait_u32
+#define emscripten_wasm_wait_i64 emscripten_atomic_wait_u64
+#define emscripten_wasm_notify emscripten_atomic_notify
+#pragma clang deprecated(emscripten_wasm_wait_i32, "use emscripten_atomic_wait_u32 instead")
+#pragma clang deprecated(emscripten_wasm_wait_i64, "use emscripten_atomic_wait_u64 instead")
+#pragma clang deprecated(emscripten_wasm_notify, "use emscripten_atomic_notify instead")
+
 #ifdef __cplusplus
 }
 #endif
