@@ -3,12 +3,18 @@
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
 # found in the LICENSE file.
 
-from . import webassembly
+import logging
+from typing import List, Dict
+
+from . import webassembly, utils
 from .webassembly import OpCode, AtomicOpCode, MemoryOpCode
 from .shared import exit_with_error
 from .settings import settings
 
 
+logger = logging.getLogger('extract_metadata')
+
+
 def skip_function_header(module):
   num_local_decls = module.read_uleb()
   while num_local_decls:
@@ -84,12 +90,12 @@ def parse_function_for_memory_inits(module, func_index, offset_map):
   call_targets = []
   while module.tell() != end:
     opcode = OpCode(module.read_byte())
-    if opcode in (OpCode.END, OpCode.NOP, OpCode.DROP, OpCode.I32_ADD):
+    if opcode in (OpCode.END, OpCode.NOP, OpCode.DROP, OpCode.I32_ADD, OpCode.I64_ADD):
       pass
     elif opcode in (OpCode.BLOCK,):
       module.read_type()
     elif opcode in (OpCode.I32_CONST, OpCode.I64_CONST):
-      const_values.append(module.read_uleb())
+      const_values.append(module.read_sleb())
     elif opcode in (OpCode.GLOBAL_SET, OpCode.BR, OpCode.GLOBAL_GET, OpCode.LOCAL_SET, OpCode.LOCAL_GET, OpCode.LOCAL_TEE):
       module.read_uleb()
     elif opcode == OpCode.CALL:
@@ -99,7 +105,7 @@ def parse_function_for_memory_inits(module, func_index, offset_map):
       if opcode == MemoryOpCode.MEMORY_INIT:
         segment_idx = module.read_uleb()
         segment = segments[segment_idx]
-        offset = const_values[-3]
+        offset = to_unsigned(const_values[-3])
         offset_map[segment] = offset
         memory = module.read_uleb()
         assert memory == 0
@@ -143,12 +149,19 @@ def get_passive_segment_offsets(module):
   return offset_map
 
 
+def to_unsigned(val):
+  if val < 0:
+    return val & ((2 ** 32) - 1)
+  else:
+    return val
+
+
 def find_segment_with_address(module, address):
   segments = module.get_segments()
   active = [s for s in segments if s.init]
 
   for seg in active:
-    offset = get_const_expr_value(seg.init)
+    offset = to_unsigned(get_const_expr_value(seg.init))
     if offset is None:
       continue
     if address >= offset and address < offset + seg.size:
@@ -177,20 +190,23 @@ def data_to_string(data):
   return data
 
 
-def get_asm_strings(module, export_map):
-  if '__start_em_asm' not in export_map or '__stop_em_asm' not in export_map:
+def get_section_strings(module, export_map, section_name):
+  start_name = f'__start_{section_name}'
+  stop_name = f'__stop_{section_name}'
+  if start_name not in export_map or stop_name not in export_map:
+    logger.debug(f'no start/stop symbols found for section: {section_name}')
     return {}
 
-  start = export_map['__start_em_asm']
-  end = export_map['__stop_em_asm']
+  start = export_map[start_name]
+  end = export_map[stop_name]
   start_global = module.get_global(start.index)
   end_global = module.get_global(end.index)
-  start_addr = get_global_value(start_global)
-  end_addr = get_global_value(end_global)
+  start_addr = to_unsigned(get_global_value(start_global))
+  end_addr = to_unsigned(get_global_value(end_global))
 
   seg = find_segment_with_address(module, start_addr)
   if not seg:
-    exit_with_error('unable to find segment starting at __start_em_asm: %s' % start_addr)
+    exit_with_error(f'unable to find segment starting at __start_{section_name}: {start_addr}')
   seg, seg_offset = seg
 
   asm_strings = {}
@@ -200,7 +216,7 @@ def get_asm_strings(module, export_map):
   end = seg_offset + size
   while str_start < end:
     str_end = data.find(b'\0', str_start)
-    asm_strings[str(start_addr - seg_offset + str_start)] = data_to_string(data[str_start:str_end])
+    asm_strings[start_addr - seg_offset + str_start] = data_to_string(data[str_start:str_end])
     str_start = str_end + 1
   return asm_strings
 
@@ -225,31 +241,45 @@ def get_main_reads_params(module, export_map):
 
 def get_named_globals(module, exports):
   named_globals = {}
+  internal_start_stop_symbols = set(['__start_em_asm', '__stop_em_asm',
+                                     '__start_em_lib_deps', '__stop_em_lib_deps',
+                                     '__em_lib_deps'])
+  internal_prefixes = ('__em_js__', '__em_lib_deps')
   for export in exports:
     if export.kind == webassembly.ExternType.GLOBAL:
-      if export.name in ('__start_em_asm', '__stop_em_asm') or export.name.startswith('__em_js__'):
+      if export.name in internal_start_stop_symbols or any(export.name.startswith(p) for p in internal_prefixes):
         continue
       g = module.get_global(export.index)
       named_globals[export.name] = str(get_global_value(g))
   return named_globals
 
 
+def get_function_exports(module):
+  rtn = {}
+  for e in module.get_exports():
+    if e.kind == webassembly.ExternType.FUNC:
+      rtn[e.name] = len(module.get_function_type(e.index).params)
+  return rtn
+
+
 def update_metadata(filename, metadata):
-  declares = []
+  imports = []
   invoke_funcs = []
-  em_js_funcs = set(metadata['emJsFuncs'])
   with webassembly.Module(filename) as module:
     for i in module.get_imports():
       if i.kind == webassembly.ExternType.FUNC:
         if i.field.startswith('invoke_'):
           invoke_funcs.append(i.field)
-        elif i.field not in em_js_funcs:
-          declares.append(i.field)
+        else:
+          imports.append(i.field)
+      elif i.kind in (webassembly.ExternType.GLOBAL, webassembly.ExternType.TAG):
+        imports.append(i.field)
 
-    exports = [e.name for e in module.get_exports() if e.kind in [webassembly.ExternType.FUNC, webassembly.ExternType.TAG]]
-  metadata['declares'] = declares
-  metadata['exports'] = exports
-  metadata['invokeFuncs'] = invoke_funcs
+    metadata.function_exports = get_function_exports(module)
+    metadata.all_exports = [utils.removeprefix(e.name, '__em_js__') for e in module.get_exports()]
+
+  metadata.imports = imports
+  metadata.invokeFuncs = invoke_funcs
 
 
 def get_string_at(module, address):
@@ -259,37 +289,51 @@ def get_string_at(module, address):
   return data_to_string(data[offset:str_end])
 
 
+class Metadata:
+  imports: List[str]
+  export: List[str]
+  asmConsts: Dict[int, str]
+  jsDeps: List[str]
+  emJsFuncs: Dict[str, str]
+  emJsFuncTypes: Dict[str, str]
+  features: List[str]
+  invokeFuncs: List[str]
+  mainReadsParams: bool
+  namedGlobals: List[str]
+
+  def __init__(self):
+    pass
+
+
 def extract_metadata(filename):
-  export_names = []
-  declares = []
+  import_names = []
   invoke_funcs = []
-  global_imports = []
   em_js_funcs = {}
+  em_js_func_types = {}
 
   with webassembly.Module(filename) as module:
     exports = module.get_exports()
     imports = module.get_imports()
 
-    for i in imports:
-      if i.kind == webassembly.ExternType.GLOBAL:
-        global_imports.append(i.field)
-
     export_map = {e.name: e for e in exports}
     for e in exports:
       if e.kind == webassembly.ExternType.GLOBAL and e.name.startswith('__em_js__'):
-        name = e.name[len('__em_js__'):]
+        name = utils.removeprefix(e.name, '__em_js__')
         globl = module.get_global(e.index)
-        string_address = get_global_value(globl)
+        string_address = to_unsigned(get_global_value(globl))
         em_js_funcs[name] = get_string_at(module, string_address)
 
     for i in imports:
       if i.kind == webassembly.ExternType.FUNC:
         if i.field.startswith('invoke_'):
           invoke_funcs.append(i.field)
-        elif i.field not in em_js_funcs:
-          declares.append(i.field)
-
-    export_names = [e.name for e in exports if e.kind in [webassembly.ExternType.FUNC, webassembly.ExternType.TAG]]
+        else:
+          if i.field in em_js_funcs:
+            types = module.get_types()
+            em_js_func_types[i.field] = types[i.type]
+          import_names.append(i.field)
+      elif i.kind in (webassembly.ExternType.GLOBAL, webassembly.ExternType.TAG):
+        import_names.append(i.field)
 
     features = module.parse_features_section()
     features = ['--enable-' + f[1] for f in features if f[0] == '+']
@@ -299,15 +343,18 @@ def extract_metadata(filename):
 
     # If main does not read its parameters, it will just be a stub that
     # calls __original_main (which has no parameters).
-    metadata = {}
-    metadata['asmConsts'] = get_asm_strings(module, export_map)
-    metadata['declares'] = declares
-    metadata['emJsFuncs'] = em_js_funcs
-    metadata['exports'] = export_names
-    metadata['features'] = features
-    metadata['globalImports'] = global_imports
-    metadata['invokeFuncs'] = invoke_funcs
-    metadata['mainReadsParams'] = get_main_reads_params(module, export_map)
-    metadata['namedGlobals'] = get_named_globals(module, exports)
+    metadata = Metadata()
+    metadata.imports = import_names
+    metadata.function_exports = get_function_exports(module)
+    metadata.all_exports = [utils.removeprefix(e.name, '__em_js__') for e in exports]
+    metadata.asmConsts = get_section_strings(module, export_map, 'em_asm')
+    metadata.jsDeps = [d for d in get_section_strings(module, export_map, 'em_lib_deps').values() if d]
+    metadata.emJsFuncs = em_js_funcs
+    metadata.emJsFuncTypes = em_js_func_types
+    metadata.features = features
+    metadata.invokeFuncs = invoke_funcs
+    metadata.mainReadsParams = get_main_reads_params(module, export_map)
+    metadata.namedGlobals = get_named_globals(module, exports)
+
     # print("Metadata parsed: " + pprint.pformat(metadata))
     return metadata
