@@ -4,8 +4,8 @@
 # found in the LICENSE file.
 
 import os
-import logging
 import re
+from tools import diagnostics
 
 TAG = 'version_3_3'
 HASH = 'd7b22660036c684f09754fcbbc7562984f02aa955eef2b76555270c63a717e6672c4fe695afb16280822e8b7c75d4b99ae21975a01a4ed51cad957f7783722cd'
@@ -18,11 +18,10 @@ def needed(settings):
 
 
 def get(ports, settings, shared):
-  ports.fetch_project(
-    'cocos2d', 'https://github.com/emscripten-ports/Cocos2d/archive/' + TAG + '.zip', 'Cocos2d-' + TAG, sha512hash=HASH)
+  ports.fetch_project('cocos2d', f'https://github.com/emscripten-ports/Cocos2d/archive/{TAG}.zip', sha512hash=HASH)
 
   def create(final):
-    logging.warn('cocos2d: library is experimental, do not expect that it will work out of the box')
+    diagnostics.warning('experimental', 'cocos2d: library is experimental, do not expect that it will work out of the box')
 
     cocos2d_src = os.path.join(ports.get_dir(), 'cocos2d')
     cocos2d_root = os.path.join(cocos2d_src, 'Cocos2d-' + TAG)
@@ -41,8 +40,8 @@ def get(ports, settings, shared):
       # See https://github.com/emscripten-ports/Cocos2d/pull/3
       '-DEMSCRIPTEN',
       '-DCP_USE_DOUBLES=0',
-      '-sUSE_ZLIB=1',
-      '-sUSE_LIBPNG=1',
+      '-sUSE_ZLIB',
+      '-sUSE_LIBPNG',
     ]
 
     for dirname in includes:
@@ -51,14 +50,15 @@ def get(ports, settings, shared):
 
     ports.build_port(cocos2d_src, final, 'cocos2d',
                      flags=flags,
+                     cxxflags=['-std=c++14'],
                      includes=includes,
                      srcs=srcs)
 
-  return [shared.Cache.get_lib('libcocos2d.a', create, what='port')]
+  return [shared.cache.get_lib('libcocos2d.a', create, what='port')]
 
 
 def clear(ports, settings, shared):
-  shared.Cache.erase_lib('libcocos2d.a')
+  shared.cache.erase_lib('libcocos2d.a')
 
 
 def process_dependencies(settings):
