@@ -22,7 +22,9 @@ import sys
 
 __scriptdir__ = os.path.dirname(os.path.abspath(__file__))
 __rootdir__ = os.path.dirname(__scriptdir__)
-sys.path.append(__rootdir__)
+sys.path.insert(0, __rootdir__)
+
+from tools import utils
 
 logger = logging.getLogger('wasm-sourcemap')
 
@@ -62,9 +64,9 @@ class Prefixes:
     for p in self.prefixes:
       if name.startswith(p['prefix']):
         if p['replacement'] is None:
-          result = name[len(p['prefix'])::]
+          result = utils.removeprefix(name, p['prefix'])
         else:
-          result = p['replacement'] + name[len(p['prefix'])::]
+          result = p['replacement'] + utils.removeprefix(name, p['prefix'])
         break
     self.cache[name] = result
     return result
@@ -249,10 +251,6 @@ def read_dwarf_entries(wasm, options):
   return sorted(entries, key=lambda entry: entry['address'])
 
 
-def normalize_path(path):
-  return path.replace('\\', '/').replace('//', '/')
-
-
 def build_sourcemap(entries, code_section_offset, prefixes, collect_sources, base_path):
   sources = []
   sources_content = [] if collect_sources else None
@@ -273,7 +271,7 @@ def build_sourcemap(entries, code_section_offset, prefixes, collect_sources, bas
       column = 1
     address = entry['address'] + code_section_offset
     file_name = entry['file']
-    file_name = normalize_path(file_name)
+    file_name = utils.normalize_path(file_name)
     # if prefixes were provided, we use that; otherwise, we emit a relative
     # path
     if prefixes.provided():
@@ -283,7 +281,7 @@ def build_sourcemap(entries, code_section_offset, prefixes, collect_sources, bas
         file_name = os.path.relpath(file_name, base_path)
       except ValueError:
         file_name = os.path.abspath(file_name)
-      file_name = normalize_path(file_name)
+      file_name = utils.normalize_path(file_name)
       source_name = file_name
     if source_name not in sources_map:
       source_id = len(sources)
